import pytest
from lark import UnexpectedToken, Tree, Token
from lark.visitors import Transformer

from eunomia.config.nodes._util_lark import SUB_PARSER, SUB_RECONSTRUCTOR


# ========================================================================= #
# Test Helper                                                               #
# ========================================================================= #


def parse(string) -> Tree:
    # TODO: this might actually be useful in the main code...

    # replace template_exp nodes by reconstructing
    # them for easier testing
    class ExprTransformer(Transformer):
        def template_exp(self, children):
            return ('exp', SUB_RECONSTRUCTOR.reconstruct(Tree('template_exp', children)))
        def template_ref(self, tokens):
            assert all(isinstance(tok, Token) for tok in tokens)
            return ('ref', '.'.join(str(tok) for tok in tokens) if tokens else None)
        def substitute_string(self, tokens):
            # this is a hack because of the grammer, just join
            # repetative chars together to form a string. so its easier to test.
            stack, str_stack = [], []
            for t in tokens:
                if isinstance(t, tuple):
                    if str_stack:
                        stack.append(''.join(str_stack))
                        str_stack.clear()
                    stack.append(t)
                else:
                    str_stack.append(t)
            if str_stack:
                stack.append(''.join(str_stack))
            return Tree('substitute_string', stack)

    # assert that the first node of the
    # tree is always 'substitute' and contains one child
    tree: Tree = SUB_PARSER.parse(string)
    assert tree.data == 'substitute'
    assert len(tree.children) == 1
    tree: Tree = tree.children[0]

    # assert that the second node of
    # the tree also only contains one child
    assert tree.data in {'substitute_string', 'interpret_fstring'}
    transformed = ExprTransformer().transform(tree)
    return transformed


# ========================================================================= #
# Test Grammar                                                              #
# ========================================================================= #


def test_lark_grammar():

    # helpers for testing parsed & transformed nodes
    def str(string):       return string
    def ref(string):       return ('ref', string)
    def exp(string):       return ('exp', string)

    def root_str(*args):   return Tree('substitute_string', list(args))
    def root_fstr(string): return Tree('interpret_fstring', [Token('FSTRING', string)])

    # parse strings
    assert parse('')                           == root_str()
    assert parse('asdf')                       == root_str(str('asdf'))
    assert parse('asdf   asdf')                == root_str(str('asdf   asdf'))
    assert parse('asdf   \t asdf')             == root_str(str('asdf   \t asdf'))
    assert parse('  \t asdf  \t  ')            == root_str(str('  \t asdf  \t  '))

    # basic reference
    assert parse('${a1.b2.c3}')  == root_str(ref('a1.b2.c3'))
    assert parse('${a1}')        == root_str(ref('a1'))
    with pytest.raises(UnexpectedToken):
        parse('${}')
    assert parse('     asdf   ${a}   ') == root_str(str('     asdf   '), ref('a'), str('   '))
    assert parse('\t    asdf   ${a}\t') == root_str(str('\t    asdf   '), ref('a'), str('\t'))

    # basic expr
    assert parse('${= 1+2}')              == root_str(exp('1+2'))
    assert parse('${=1 + 2}')             == root_str(exp('1+2'))
    assert parse('${=1 - 1}')             == root_str(exp('1-1'))
    with pytest.raises(UnexpectedToken):
        parse('${=}')

    # basic fstr
    assert parse('f"1+2+3"')              == root_fstr('f"1+2+3"')
    assert parse('f"1+2+{\'asdf\'}+3"')   == root_fstr('f"1+2+{\'asdf\'}+3"')
    assert parse("f'1+2+3={1+2+3}'")      == root_fstr("f'1+2+3={1+2+3}'")
    assert parse("f'1+2+3=${=1+2}'")      == root_fstr("f'1+2+3=${=1+2}'")

    # more complex combined
    assert parse('${a1}${=+1}${b2}')                  == root_str(ref('a1'), exp('+1'), ref('b2'))
    assert parse('asdf${a1}${=+1}${b2}')              == root_str(str('asdf'), ref('a1'), exp('+1'), ref('b2'))
    assert parse('asdf${a.b.c}fdaasd${a}${=1+2}')     == root_str(str('asdf'), ref('a.b.c'), str('fdaasd'), ref('a'), exp('1+2'))
    with pytest.raises(UnexpectedToken):
        parse('as df${a.b.c}fdaa\nsd${a}${{1+2}}')

    # these have a tendency to be buggy
    assert parse('${=[1]}')             == root_str(exp('[1]'))
    assert parse('${={1}}')             == root_str(exp('{1,}'))
    assert parse('${={ 1 }}')           == root_str(exp('{1,}'))
    assert parse('${={}}')              == root_str(exp('{}'))
    assert parse('${={1: 2}}')          == root_str(exp('{1:2,}'))
    assert parse('${={1, 2, 3}}')       == root_str(exp('{1,2,3,}'))
    assert parse('${={1,2}|{3,4}}')     == root_str(exp('{1,2,}|{3,4,}'))

    # test semi-colons
    with pytest.raises(UnexpectedToken): parse('${=;}')
    # disabled colons -- otherwise enable these
    with pytest.raises(UnexpectedToken): assert parse('${=1+2;}')            == root_str(exp('1+2'))
    with pytest.raises(UnexpectedToken): assert parse('as df${a.b.c}fdaa\tsd${a}${=1+2;}') == root_str(str('as df'), ref('a.b.c'), str('fdaa\tsd'), ref('a'), exp('1+2'))
    with pytest.raises(UnexpectedToken): assert parse('${=(1,);}')           == root_str(exp('(1,)'))
    with pytest.raises(UnexpectedToken): assert parse('${={1:2} ;}')         == root_str(exp('{1:2,}'))
    with pytest.raises(UnexpectedToken): assert parse('${= {1: 2} ;}')       == root_str(exp('{1:2,}'))
    with pytest.raises(UnexpectedToken): assert parse('${={};}')             == root_str(exp('{}'))
    with pytest.raises(UnexpectedToken): assert parse('f"asdf";')            == root_fstr('f"asdf"')

    # # check contained brackets
    assert parse('{}') == root_str(str('{}'))
    assert parse('a{}b') == root_str(str('a{}b'))
    assert parse('a{123}b') == root_str(str('a{123}b'))
    assert parse('a{123;}b') == root_str(str('a{123;}b'))
    assert parse('a{=123}b') == root_str(str('a{=123}b'))
    assert parse('a{=123;}b') == root_str(str('a{=123;}b'))
    assert parse('{=}') == root_str(str('{=}'))
    assert parse('{=1}') == root_str(str('{=1}'))
    assert parse('a{{}}b') == root_str(str('a{{}}b'))
    assert parse('a{{};}b') == root_str(str('a{{};}b'))
    assert parse('a{={}}b') == root_str(str('a{={}}b'))
    assert parse('a{={};}b') == root_str(str('a{={};}b'))



import pytest

from eunomia.backend import BackendObj, BackendDict, BackendYaml
from tests.test_backend_obj import _make_config_group


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def test_simple_option():
    root = _make_config_group()

    b_obj = BackendObj()
    b_dct = BackendDict()

    # test group backend
    assert b_dct.dump(b_obj.load_group(root)) == b_dct.dump(root)
    with pytest.raises(TypeError, match='group value must be of type: .*Group'):
        b_dct.dump(b_obj.load_group(b_dct.dump(root)))

    # test dict backend
    assert b_dct.dump(b_dct.load_group(b_dct.dump(root))) == b_dct.dump(root)
    with pytest.raises(TypeError, match='group value must be of type: .*dict'):
        b_dct.dump(b_dct.load_group(root))

    # test yaml backend
    BackendYaml().load_group('examples/docs/quickstart/configs')


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



import pytest

from eunomia import eunomia, eunomia_runner, eunomia_load
from eunomia.backend import BackendObj, BackendDict
from eunomia.config.nodes import SubNode, EvalNode
from eunomia.config.nodes._nodes import OptNode
from tests.test_backend_obj import _make_config_group


# ========================================================================= #
# Test Eunomia                                                              #
# ========================================================================= #


def test_eunomia_loader_simple():
    eunomia_load(_make_config_group(suboption='suboption1', suboption2=None))
    eunomia_load(_make_config_group(suboption='suboption2', suboption2=None))
    eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption1'))
    eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption2'))
    eunomia_load(_make_config_group(suboption=None, suboption2='suboption1'))
    eunomia_load(_make_config_group(suboption=None, suboption2='suboption2'))


def test_eunomia_loader():
    # check no subgroups
    assert eunomia_load(_make_config_group(suboption=None)) == {'foo': 1}

    # test subgroup
    assert eunomia_load(_make_config_group(suboption='suboption1')) == {'subgroup': {'bar': 1}, 'foo': 1}
    assert eunomia_load(_make_config_group(suboption='suboption2')) == {'subgroup': {'bar': 2}, 'foo': 1}

    with pytest.raises(Exception):
        eunomia_load(_make_config_group(suboption='invalid___'))

    # test second subgroup
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption1')) == {'foo': 1, 'subgroup': {'bar': 2}, 'subgroup2': {'subgroup3': {'baz': 1}}}
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption2')) == {'foo': 1, 'subgroup': {'bar': 2}, 'subgroup2': {'subgroup3': {'baz': 2}}}

    # test root package
    assert eunomia_load(_make_config_group(suboption2='suboption2', package1='<root>')) == {'foo': 1, 'bar': 1, 'subgroup2': {'subgroup3': {'baz': 2}}}
    assert eunomia_load(_make_config_group(suboption2='suboption2', package2='<root>')) == {'foo': 1, 'subgroup': {'bar': 1}, 'baz': 2}

    # test package variations
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption2', package1='<root>', package2='<root>')) == {'foo': 1, 'bar': 2, 'baz': 2}
    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption2', package1='<root>', package2='<group>')) == {'foo': 1, 'bar': 1, 'subgroup2': {'subgroup3': {'baz': 2}}}
    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption1', package1='<root>', package2='<option>')) == {'foo': 1, 'bar': 1, 'subgroup2': {'subgroup3': {'suboption1': {'baz': 1}}}}

    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption1', package1='<group>', package2='<root>')) == {'foo': 1, 'subgroup': {'bar': 1}, 'baz': 1}
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption2', package1='<group>', package2='<group>')) == {'foo': 1, 'subgroup': {'bar': 2}, 'subgroup2': {'subgroup3': {'baz': 2}}}
    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption2', package1='<group>', package2='<option>')) == {'foo': 1, 'subgroup': {'bar': 1}, 'subgroup2': {'subgroup3': {'suboption2': {'baz': 2}}}}

    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption2', package1='<option>', package2='<root>')) == {'foo': 1, 'subgroup': {'suboption1': {'bar': 1}}, 'baz': 2}
    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2='suboption1', package1='<option>', package2='<group>')) == {'foo': 1, 'subgroup': {'suboption1': {'bar': 1}}, 'subgroup2': {'subgroup3': {'baz': 1}}}
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption2', package1='<option>', package2='<option>')) == {'foo': 1, 'subgroup': {'suboption2': {'bar': 2}}, 'subgroup2': {'subgroup3': {'suboption2': {'baz': 2}}}}

    # test custom package
    assert eunomia_load(_make_config_group(suboption2='suboption2', package1='asdf', package2='fdsa.asdf')) == {'asdf': {'bar': 1}, 'fdsa': {'asdf': {'baz': 2}}, 'foo': 1}
    assert eunomia_load(_make_config_group(suboption2='suboption2', package1='asdf', package2='asdf.fdsa')) == {'asdf': {'bar': 1, 'fdsa': {'baz': 2}}, 'foo': 1}


def test_eunomia_loader_overrides():
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption1')) == {'foo': 1, 'subgroup': {'bar': 2}, 'subgroup2': {'subgroup3': {'baz': 1}}}
    # change suboption2=suboption2
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption1'), overrides=['/subgroup2/subgroup3/suboption2']) == {'foo': 1, 'subgroup': {'bar': 2}, 'subgroup2': {'subgroup3': {'baz': 2}}}
    # change suboption=suboption1 and suboption2=suboption2
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption1'), overrides=['/subgroup2/subgroup3/suboption2', '/subgroup/suboption1']) == {'foo': 1, 'subgroup': {'bar': 1}, 'subgroup2': {'subgroup3': {'baz': 2}}}
    # check missing override
    with pytest.raises(KeyError, match='could not resolve the override'):
        eunomia_load(_make_config_group(suboption='suboption2', suboption2='suboption1'), overrides=['/subgroup2/subgroup3/suboption2', '/MALFORMED/suboption1'])
    # check unused override
    with pytest.raises(RuntimeError, match="the following overrides were not used to override defaults listed in the config: \\['/subgroup2/subgroup3']"):
        eunomia_load(_make_config_group(suboption='suboption2', suboption2=None), overrides=['/subgroup2/subgroup3/suboption2', '/subgroup/suboption1'])
    with pytest.raises(RuntimeError, match="the following overrides were not used to override defaults listed in the config: \\['/subgroup', '/subgroup2/subgroup3']"):
        eunomia_load(_make_config_group(suboption=None, suboption2=None), overrides=['/subgroup2/subgroup3/suboption2', '/subgroup/suboption1'])


def test_eunomia_loader_interpolation():
    # test custom packages with substitution
    with pytest.raises(TypeError, match='can never be a config node'):
        assert eunomia_load(_make_config_group(suboption2='suboption2', package1='asdf', package2=SubNode('asdf.fdsa')))

    # test interpolation values
    assert eunomia_load(_make_config_group(suboption=SubNode('suboption${=1}'))) == {'subgroup': {'bar': 1}, 'foo': 1}
    assert eunomia_load(_make_config_group(suboption=SubNode('f"suboption{1}"'))) == {'subgroup': {'bar': 1}, 'foo': 1}
    assert eunomia_load(_make_config_group(suboption=EvalNode('f"suboption{2}"'))) == {'subgroup': {'bar': 2}, 'foo': 1}
    assert eunomia_load(_make_config_group(suboption=SubNode('suboption${foo}'))) == {'subgroup': {'bar': 1}, 'foo': 1}

    # test option name references
    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2=OptNode('/subgroup'))) == {'subgroup': {'bar': 1}, 'foo': 1, 'subgroup2': {'subgroup3': {'baz': 1}}}
    assert eunomia_load(_make_config_group(suboption='suboption1', suboption2=SubNode('${/subgroup}'))) == {'subgroup': {'bar': 1}, 'foo': 1, 'subgroup2': {'subgroup3': {'baz': 1}}}
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2=OptNode('/subgroup'))) == {'subgroup': {'bar': 2}, 'foo': 1, 'subgroup2': {'subgroup3': {'baz': 2}}}
    assert eunomia_load(_make_config_group(suboption='suboption2', suboption2=SubNode('${/subgroup}'))) == {'subgroup': {'bar': 2}, 'foo': 1, 'subgroup2': {'subgroup3': {'baz': 2}}}
    with pytest.raises(KeyError):
        eunomia_load(_make_config_group(suboption=SubNode('${/subgroup}'), suboption2='suboption2'))


def test_eunomia_core_funcs():

    root = _make_config_group()
    target = {'foo': 1, 'subgroup': {'bar': 1}}

    # LOADER

    # test group backend
    assert eunomia_load(root, 'default') == target
    assert eunomia_load(root, 'default', backend=BackendObj()) == target

    # test dict backend
    assert eunomia_load(BackendDict().dump(root), 'default') == target
    assert eunomia_load(BackendDict().dump(root), 'default', backend=BackendDict()) == target

    # RUNNERS

    def main(config):
        assert config == target
    eunomia_runner(main, root, 'default')
    eunomia_runner(main, root, 'default', backend=BackendObj())

    # WRAPPERS

    @eunomia(root, 'default')
    def main(config):
        assert config == target
    main()

    @eunomia(root, 'default', backend=BackendObj())
    def main(config):
        assert config == target
    main()





import pytest

from tests.util import temp_capture_stdout, temp_wd, to_import_path, path_from_root, read_from_root

TESTS = [
    ('examples/docs/quickstart', 'simple_yaml'),
    ('examples/docs/quickstart', 'simple_yaml_alt'),
    ('examples/docs/quickstart', 'simple_pythonic'),
    ('examples/docs/quickstart', 'advanced_yaml'),
]


@pytest.mark.parametrize(["folder", "name"], TESTS)
def test_docs_examples_eval_to_targets(folder, name):
    # get test files
    import_path = to_import_path(path_from_root(folder, f'{name}.py'))
    target_file = path_from_root(folder, f'{name}.target')
    # run all the files in the examples folder
    import importlib
    with temp_wd(folder):
        with temp_capture_stdout() as f:
            importlib.import_module(import_path)
    assert f.getvalue() == read_from_root(target_file)



from typing import List as _List


from eunomia.config import ConfigLoader
from eunomia.backend import Backend, BackendObj, BackendYaml, BackendDict
from eunomia.backend import DefaultConfigTypes as _ValidConfigTypes, infer_backend_load_group as _infer_backend_load_group


# ========================================================================= #
# Variables                                                                 #
# ========================================================================= #


# default config gives BackedYaml pointing to ./configs
# default entrypoint recursively loads ./configs/default.yaml
DEFAULT_CONFIG = 'configs'
DEFAULT_ENTRYPOINT = 'default'


# ========================================================================= #
# Decorators                                                                #
# ========================================================================= #


def eunomia(
        config: _ValidConfigTypes = DEFAULT_CONFIG,
        entrypoint=DEFAULT_ENTRYPOINT,
        backend: Backend = None,
):
    """
    The main eunomia decorator.
    Automatically detects which backend to use based on the first argument.
    """
    def wrapper(func):
        from functools import wraps
        @wraps(func)
        def runner():
            return eunomia_runner(func=func, config=config, entrypoint=entrypoint, backend=backend)
        return runner
    return wrapper


def eunomia_runner(
        func: callable,
        config: _ValidConfigTypes = DEFAULT_CONFIG,
        entrypoint=DEFAULT_ENTRYPOINT,
        overrides: _List[str] = None,
        backend: Backend = None,
):
    """
    The non-decorator equivalent to @eunomia(...)
    - This function is the core of eunomia, calling the relevant plugins, creating
      the merged config and finally calling your entry.
    """
    config = eunomia_load(
        config=config,
        entrypoint=entrypoint,
        backend=backend,
    )
    # TODO: extract runner from config, and run that way!
    func(config)


# ========================================================================= #
# Single Config Loading - No Wrapping                                       #
# ========================================================================= #


def eunomia_load(
        config: _ValidConfigTypes = DEFAULT_CONFIG,
        entrypoint=DEFAULT_ENTRYPOINT,
        overrides: _List[str] = None,
        backend: Backend = None,
):
    group = _infer_backend_load_group(config, backend=backend)
    loader = ConfigLoader(group, overrides=overrides)
    return loader.load_config(entrypoint)


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



from copy import deepcopy

import pytest
from eunomia.util import _util_dict as u


def test_dict_util():
    _left = dict(a=1, b=dict(a=10, b=20))

    left = deepcopy(_left)
    assert u.recursive_getitem(left, ['a']) == 1
    assert u.recursive_getitem(left, ['b', 'a']) == 10
    assert u.recursive_getitem(left, []) is left
    assert u.recursive_getitem(left, ['b']) is left['b']

    # set values
    left = deepcopy(_left)
    u.recursive_setitem(left, ['a'], 10)
    assert left['a'] == 10
    u.recursive_setitem(left, ['b', 'a'], 22)
    assert left['b']['a'] == 22

    # missing keys
    left = deepcopy(_left)
    u.recursive_setitem(left, ['c'], 42)  # alright because its the last one
    u.recursive_setitem(left, ['b', 'c'], 42)  # alright because its the last one
    with pytest.raises(KeyError, match='d'):
        u.recursive_setitem(left, ['d', 'a'], 42)
    u.recursive_setitem(left, ['d', 'a'], 42, make_missing=True)
    assert left == dict(a=1, b=dict(a=10, b=20, c=42), c=42, d=dict(a=42))
    with pytest.raises(KeyError, match='requires as least one key'):
        u.recursive_setitem(left, [], 42)

    # overwrite
    left = deepcopy(_left)
    u.recursive_setitem(left, ['b'], 42)
    assert left == dict(a=1, b=42)

    # update
    left = deepcopy(_left)
    u.dict_recursive_update(left, right=dict(c=42))
    assert left == dict(a=1, b=dict(a=10, b=20), c=42)

    # update -- both dictionaries
    left = deepcopy(_left)
    u.dict_recursive_update(left, right=dict(b=dict(a=42)))
    assert left == dict(a=1, b=dict(a=42, b=20))

    # update -- different types
    left = deepcopy(_left)
    with pytest.raises(TypeError, match='cannot update keys with different types'):
        u.dict_recursive_update(left, right=dict(b=42))

    left = deepcopy(_left)
    with pytest.raises(TypeError, match='cannot update keys with different types'):
        u.dict_recursive_update(left, right=dict(b=42), safe_merge=True)

    left = deepcopy(_left)
    u.dict_recursive_update(left, right=dict(b=42), safe_merge=False)
    assert left == dict(a=1, b=42)

    left = deepcopy(_left)
    u.dict_recursive_update(left, right=dict(b=42), safe_merge=True, allow_update_types=[(dict, int)])
    assert left == dict(a=1, b=42)

    # floats and integers should be allowed together, this is a special case

    _left = dict(a=1, b=7.7, c=(1,2,3))

    left = deepcopy(_left)
    with pytest.raises(TypeError, match='cannot update keys with different types'):
        u.dict_recursive_update(left, right=dict(c=42))
    left = deepcopy(_left)
    u.dict_recursive_update(left, right=dict(b=42))

    # MAYBE?
    # lists and tuples should be allowed together, this is a special case
    # lists replace lists
    # left = deepcopy(_left)
    # u.dict_recursive_update(left, right=dict(c=[2,3,4]))



from ._backend import Backend
from ._backend_dict import BackendDict
from ._backend_obj import BackendObj
from ._backend_yaml import BackendYaml

from ._helper import infer_backend_load_group, DefaultConfigTypes



from typing import Any, Union, List, Tuple
import lark
from eunomia.config.nodes._util_interpret import interpret_expr
from eunomia.config.nodes._util_lark import SUB_RECONSTRUCTOR, SUB_PARSER
from eunomia.config import validate as V



# ========================================================================= #
# Base Loaders                                                              #
# ========================================================================= #


class ConfigNode(object):

    @property
    def INSTANCE_OF(self) -> Union[type, Tuple[type, ...]]:
        raise NotImplementedError

    def __init__(self, value: str):
        assert isinstance(value, self.INSTANCE_OF), f'value={repr(value)} corresponding to {self.__class__.__name__} must be instance of: {self.INSTANCE_OF}'
        self.raw_value = value

    def __repr__(self):
        return f'{self.__class__.__name__}(value={repr(self.raw_value)})'

    def __str__(self):
        return repr(self)

    def get_config_value(self, merged_config: dict, merged_options: dict, current_config: dict):
        raise NotImplementedError

    def __eq__(self, other):
        if not isinstance(other, self.__class__): return False
        if not isinstance(self, other.__class__): return False
        return other.raw_value == self.raw_value

    def __hash__(self):
        return hash((self.__class__, self.raw_value))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Recursive Resolve                                                     #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @staticmethod
    def recursive_get_config_value(merged_config: dict, merged_options: dict, current_config: dict, value: Any):
        if isinstance(value, ConfigNode):
            return value.get_config_value(merged_config, merged_options, current_config)
        elif isinstance(value, list):
            return list(ConfigNode.recursive_get_config_value(merged_config, merged_options, current_config, v) for v in value)
        elif isinstance(value, tuple):
            return tuple(ConfigNode.recursive_get_config_value(merged_config, merged_options, current_config, v) for v in value)
        elif isinstance(value, set):
            return set(ConfigNode.recursive_get_config_value(merged_config, merged_options, current_config, v) for v in value)
        elif isinstance(value, dict):
            return {
                ConfigNode.recursive_get_config_value(merged_config, merged_options, current_config, k):
                    ConfigNode.recursive_get_config_value(merged_config, merged_options, current_config, v)
                for k, v in value.items()
            }
        else:
            return value


# ========================================================================= #
# Basic Nodes                                                               #
# ========================================================================= #


class IgnoreNode(ConfigNode):

    INSTANCE_OF = str

    def get_config_value(self, merged_config: dict, merged_options: dict, current_config: dict):
        return self.raw_value

    def __str__(self):
        return self.raw_value


class RefNode(ConfigNode):

    INSTANCE_OF = str

    def get_config_value(self, merged_config: dict, merged_options: dict, current_config: dict) -> Any:
        keys, is_relative = V.split_package_path(self.raw_value)
        if is_relative:
            raise ValueError(f'reference cannot be a relative path, must be from root: {self.raw_value}')
        # walk to get value
        value = merged_config
        for key in keys:
            value = value[key]
            # resolve
            if isinstance(value, ConfigNode):
                value = value.get_config_value(merged_config, merged_options, current_config)
        return value

    def __str__(self):
        return f'${{{self.raw_value}}}'


class OptNode(ConfigNode):
    """
    Get the name of a chosen option from the path to a group.
    """

    INSTANCE_OF = str

    def get_config_value(self, merged_config: dict, merged_options: dict, current_config: dict) -> Any:
        keys, is_relative = V.split_config_path(self.raw_value)
        if is_relative:
            raise ValueError(f'option choice cannot be a relative path, must be from root: {self.raw_value}')
        # check that a choice has been made so far!
        keys, path = tuple(keys), "/" + "/".join(keys)
        if keys not in merged_options:
            raise KeyError(f'No group choice or default has yet been merged into the config for: {path}\nAre you sure your defaults are ordered correctly or that you did not specify the path to an option itself, the path must be to a group?')
        # get the name
        return merged_options[keys][-1]

    def __str__(self):
        return f'${{{self.raw_value}}}'


class EvalNode(ConfigNode):

    INSTANCE_OF = str

    def get_config_value(self, merged_config: dict, merged_options: dict, current_config: dict):
        return interpret_expr(
            self.raw_value,
            usersyms={
                'this': current_config,
                'conf': merged_config,
                'incl': merged_options,
            },
            NON_STANDARD_PYTHON=True  # try getitem on AttributeError
        )

    def __str__(self):
        return f'${{={self.raw_value}}}'


# ========================================================================= #
# Substitute Nodes                                                          #
# ========================================================================= #


class SubNode(ConfigNode):

    INSTANCE_OF = (str, list)
    ALLOWED_SUB_NODES = (str, IgnoreNode, RefNode, OptNode, EvalNode)

    def _check_subnodes(self, nodes: list):
        for subnode in nodes:
            if not isinstance(subnode, self.ALLOWED_SUB_NODES):
                raise TypeError(f'Malformed {SubNode.__name__}, subnode={repr(subnode)} must be instance of: {self.ALLOWED_SUB_NODES}')

    def get_config_value(self, merged_config: dict, merged_options: dict, current_config: dict) -> str:
        nodes = self.raw_value

        # 1. convert string to nodes if necessary using lark
        # detects f-strings f"..." or f'...' and strings
        # with placeholders ${...} and ${=...} defined in
        # the lark grammar
        if isinstance(nodes, str):
            nodes = _string_to_sub_nodes(nodes)
        self._check_subnodes(nodes)

        # 2. concatenate strings and substituted values
        # obtained from calling this same function on
        # nodes and child nodes
        values = []
        for subnode in nodes:
            if not isinstance(subnode, str):
                subnode = ConfigNode.recursive_get_config_value(merged_config, merged_options, current_config, subnode)
            values.append(subnode)

        # 3. get final result -- return that actual value if its the
        # only value in the list, instead of merging to a string
        if len(values) == 1:
            return values[0]
        return ''.join(str(v) for v in values)

    def __str__(self):
        if isinstance(self.raw_value, str):
            return self.raw_value
        else:
            return ''.join(map(str, self.raw_value))


def _string_to_sub_nodes(string):
    nodes = SUB_PARSER.parse(string)
    converted = _InterpretLarkToConfNodesList().visit(nodes)
    return converted


class _InterpretLarkToConfNodesList(lark.visitors.Interpreter):
    """
    This class walks a lark.Tree and converts the lark nodes as
    necessary to action nodes.
    - RefNode
    - EvalNode
    - IgnoreNode

    The lark nodes are defined in the interpolation grammar file.
    grammar_substitute.lark
    """

    def substitute(self, tree) -> List[ConfigNode]:
        if len(tree.children) != 1:
            raise RuntimeError('Malformed substitute node. This should never happen!')
        return self.visit(tree.children[0])

    def substitute_string(self, tree) -> List[ConfigNode]:
        return self.visit_children(tree)

    def interpret_fstring(self, tree) -> List[ConfigNode]:
        # TODO: having to wrap in a list here might be a grammar mistake
        return [
            EvalNode(SUB_RECONSTRUCTOR.reconstruct(tree))
        ]

    def template_ref(self, tree): return RefNode(SUB_RECONSTRUCTOR.reconstruct(tree))
    def template_opt(self, tree): return OptNode(SUB_RECONSTRUCTOR.reconstruct(tree))
    def template_exp(self, tree): return EvalNode(SUB_RECONSTRUCTOR.reconstruct(tree))

    # we no longer handle this due to the grammar change for CHAR and WSSTR
    # def str(self, node):        return IgnoreNode(SUB_RECONSTRUCTOR.reconstruct(node))

    def __getattr__(self, item):
        raise RuntimeError(f'This should never happen! Unknown Interpolation Grammar Node Visited: {item}')


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



from typing import Union, Dict, List, Tuple

from eunomia.config import Option, Group
from eunomia.config import keys as K
from eunomia.config import validate as V
from eunomia.config.nodes import SubNode, ConfigNode
from eunomia.util._util_traverse import Transformer


# ========================================================================= #
# Defaults Entry                                                            #
# ========================================================================= #


OorS = Union[Group, Option, str]
GorS = Union[Group, Option, str]
GorOorS = Union[Group, Option, str]

InputDefaultType = Union[
    'Default',
    # we can infer lhs, this is the rhs
    # -- this is converted to below
    GorOorS,
    # lhs must be direct parent of rhs, more verbose
    Dict[GorOorS, Union[GorOorS, List[OorS]]],
]


# ========================================================================= #
# Defaults Entry                                                            #
# ========================================================================= #


class DefaultTransformer(Transformer):

    # TODO: this could be cleaned up a lot...
    #       very manual at the moment...

    def __init__(self, in_opt: Option, resolver, force_substitute=True):
        self._in_opt = in_opt
        self._resolver = resolver
        self._force_resolve = force_substitute

    def __transform_default__(self, value):
        if isinstance(value, ConfigNode):
            return self._transform_str(value)
        raise TypeError(f'unsupported type trying to transform default item: {value}')

    def _transform_str(self, string):
        # handle special value
        if string == K.OPT_SELF:
            raise RuntimeError(f'this is a bug, {K.OPT_SELF} should be handled before being transformed!')
        if string in K.ALL_OPT_ALIASES:
            raise RuntimeError(f'{repr(string)} is not allowed as a single string in defaults. Must be a value after a group.')
        object = self._resolve(string, from_group=self._in_opt.group)
        # this check should never fail
        assert isinstance(object, (Option, Group))
        # continue transforming
        return self.transform(object)

    def _transform_Group(self, group: Group):
        return group, list(group.options.values())

    def _transform_Option(self, option: Option):
        return option.group, [option]

    def _transform_dict(self, dct: dict):
        # checks
        if len(dct) != 1:
            raise ValueError(f'dict default entry {dct} must contain only one key.')
        ((group, opts),) = dct.items()
        assert not isinstance(group, dict)
        assert not isinstance(opts, dict)
        # resolve keys & handle glob
        if isinstance(group, str):
            group = self._resolve(group, from_group=self._in_opt.group)
        if not isinstance(group, Group):
            raise KeyError(f'defaults entry has key that is not a group or path to a group: {group.abs_path}')
        # resolve values
        if isinstance(opts, (str, ConfigNode)):
            opts = group if (opts == K.OPT_GLOB) else self._resolve(opts, from_group=group)
            assert isinstance(opts, (Group, Option))
        if isinstance(opts, (Group, Option)):
            _, opts = self.transform(opts)
            assert isinstance(opts, list)
        # checks
        assert isinstance(group, (Group, Option))
        assert isinstance(opts, list)
        group: Union[Group, Option]
        # continue resolving,
        options: List[Option] = []
        for opt in opts:
            if isinstance(opt, (str, ConfigNode)):
                opt = self._resolve(opt, from_group=group)
            if not isinstance(opt, Option):
                raise KeyError(f'defaults entry has value in list that is not an option {repr(opt.abs_path)}. All values given in a list are expected to point to options!')
            options.append(opt)
        # return finally!
        return group, options

    def _resolve(self, string, from_group: Group) -> Union[Group, Option]:
        if string in K.ALL_OPT_ALIASES:
            raise RuntimeError(f'{string} cannot be nested in a defaults entry!')
        # resolve
        if self._force_resolve:
            string = string if isinstance(string, ConfigNode) else SubNode(string)
        if isinstance(string, ConfigNode):
            if self._resolver:
                string = self._resolver(string)
        # checks
        if not isinstance(string, str):
            raise TypeError(f'default string did not resolve as a string! Invalid: {repr(string)}')
        if string in K.ALL_OPT_ALIASES:
            raise RuntimeError(f'defaults are not allowed to resolve to {string}! Invalid: {repr(string)}')
        # get config object
        return from_group.get_child_recursive(string)


class Default(object):

    def __init__(self, default: InputDefaultType, pkg_override=None):
        self._default = default
        # check override
        self._pkg_override = pkg_override
        if pkg_override is not None:
            if not isinstance(pkg_override, str):
                raise TypeError(f'pkg_override must be a string. Invalid: {pkg_override}')
            V.split_package_path(pkg_override)

    @staticmethod
    def make_self():
        return Default(K.OPT_SELF)

    @property
    def is_self_pre_resolve(self):
        return self._default == K.OPT_SELF

    def to_resolved_components(self, in_opt: Option, resolver, force_substitute=True) -> Tuple[Group, List[Option], Tuple[str, ...], bool]:
        if self.is_self_pre_resolve:
            is_self = True
            # get values
            parent, children = in_opt, [in_opt]
        else:
            is_self = False
            # get values
            parent, children = DefaultTransformer(in_opt, resolver, force_substitute=force_substitute).transform(self._default)
            # check parents and children
            assert isinstance(parent, (Group, Option))
            assert all(isinstance(child, Option) for child in children)
            assert all((child.parent is parent) for child in children)
            assert len(set(child.abs_path for child in children)) == len(children)
            # get other values
        # resolve the package
        pkg_keys = parent.get_package_keys(self._pkg_override)
        # return the values!
        return parent, children, pkg_keys, is_self

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self._default)})'



# ========================================================================= #
# End                                                                       #
# ========================================================================= #



from eunomia.backend import Backend
from eunomia.config import Group, Option

from eunomia.config import keys as K
from eunomia.config import validate as V


# ========================================================================= #
# Config Backend                                                            #
# ========================================================================= #
from eunomia.config.nodes import ConfigNode


class BackendDict(Backend):

    GROUP_TYPE = dict
    OPTION_TYPE = dict

    def __init__(self, allow_compact_load=False):
        self._allow_compact_load = allow_compact_load

    def _load_group(self, value) -> Group:
        # loads below also normalise
        value = normalise_group_dict(value, recursive=False)
        group = Group()
        for key, child in value[K.KEY_CHILDREN].items():
            if child[K.KEY_TYPE] == K.TYPE_GROUP:
                group.add_subgroup(key, self.load_group(child))
            elif child[K.KEY_TYPE] == K.TYPE_OPTION:
                group.add_option(key, self.load_option(child))
            else:
                raise ValueError(f'Invalid type: {child[K.KEY_TYPE]}')
        return group

    def _load_option(self, value) -> Option:
        # TODO: this should be recursive
        value = normalise_option_dict(value, allow_compact=self._allow_compact_load)
        return Option(
            pkg=value[K.KEY_PKG],
            defaults=value[K.KEY_DEFAULTS],
            data=value[K.KEY_DATA],
        )

    def _dump_group(self, group: Group):
        group = {
            K.KEY_TYPE: K.TYPE_GROUP,
            K.KEY_CHILDREN: {
                k: self.dump(child)
                for k, child in group._children.items()
            }
        }
        # dump normalises above
        return normalise_group_dict(group, recursive=False)

    def _dump_option(self, option: Option):
        option = {
            K.KEY_TYPE: K.TYPE_OPTION,
            K.KEY_PKG: option.pkg,
            K.KEY_DEFAULTS: option.defaults,
            K.KEY_DATA: option.data,
        }
        # TODO: this should be recursive
        return normalise_option_dict(option, allow_compact=False)


# ========================================================================= #
# VERBOSE VERSIONS                                                          #
# - these versions are used internally                                      #
# ========================================================================= #


def normalise_group_dict(group: dict, recursive: bool):
    assert not recursive, 'recursive normalise is not supported'  # does not validate children
    if not isinstance(group, dict):
        raise TypeError('group must be a dictionary')
    # ========================= #
    # check the keys
    keys = set(group.keys()) - K.RESERVED_GROUP_KEYS
    if keys:
        raise ValueError(f'keys not allowed found in the group: {keys} only these keys are allowed: {K.RESERVED_GROUP_KEYS}')
    # ========================= #
    # set defaults
    type = group.setdefault(K.KEY_TYPE, K.TYPE_GROUP)
    children = group.setdefault(K.KEY_CHILDREN, {})
    # ========================= #
    # validate values
    if not (type == K.TYPE_GROUP):
        raise ValueError(f'group {repr(K.KEY_TYPE)} must be: {repr(K.TYPE_GROUP)}')
    if not isinstance(children, dict):
        raise TypeError(f'group {repr(K.KEY_CHILDREN)} must be of type: {dict}')
    # ========================= #
    return group


def normalise_option_dict(option: dict, allow_compact=False, allow_config_nodes=False):
    if not isinstance(option, dict):
        raise TypeError('option must be a dictionary')
    # ========================= #
    # handle compact dictionary -- identified if extra keys exist
    extra_keys = set(option.keys()) - K.RESERVED_OPTION_KEYS
    if extra_keys:
        # COMPACT
        if not allow_compact:
            raise KeyError(f'compact options are not allowed, option requires key: {K.KEY_DATA}')
        if K.KEY_DATA in option:
            raise ValueError(f'option is considered compact because it has extra non-special keys in it. A compact option cannot contain a {K.KEY_DATA} key itself.')
        option = {
            K.KEY_TYPE:     option.pop(K.KEY_TYPE, None),
            K.KEY_PKG:      option.pop(K.KEY_PKG, None),
            K.KEY_DEFAULTS: option.pop(K.KEY_DEFAULTS, None),
            K.KEY_DATA:     option,
        }
    # ========================= #
    # get defaults
    option[K.KEY_TYPE]     = V.validate_option_type(option.get(K.KEY_TYPE, None))
    option[K.KEY_PKG]      = V.validate_option_package(option.get(K.KEY_PKG, None))
    option[K.KEY_DEFAULTS] = V.validate_option_defaults(option.get(K.KEY_DEFAULTS, None), allow_config_nodes=allow_config_nodes)
    option[K.KEY_DATA]     = V.validate_option_data(option.get(K.KEY_DATA, None))
    # ========================= #
    # again check that no extra keys exist
    extra_keys = set(option.keys()) - K.RESERVED_OPTION_KEYS
    if extra_keys:
        raise ValueError(f'option has extra keys that are not allowed: {extra_keys}')
    # check that nothing special is in here
    extra_data_keys = set(option[K.KEY_DATA].keys()).intersection(K.RESERVED_KEYS)
    if extra_data_keys:
        raise ValueError(f'option data has reserved keys that are not allowed: {extra_data_keys}')
    # ========================= #
    return option


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



"""
Config nodes are used to represent actions that should be
performed during the merging of the config itself.
"""

from ._nodes import ConfigNode
from ._nodes import IgnoreNode, RefNode, EvalNode, SubNode



import os
import sys
from contextlib import contextmanager
from io import StringIO


# @contextmanager
# def temp_no_stdout():
#     old_stdout = sys.stdout
#     sys.stdout = open(os.devnull, 'w')
#     yield
#     sys.stdout = old_stdout


@contextmanager
def temp_capture_stdout() -> StringIO:
    old_stdout, temp_out = sys.stdout, StringIO()
    sys.stdout = temp_out
    yield temp_out
    sys.stdout = old_stdout


class temp_wd:
    def __init__(self, new_wd):
        self._new_wd = os.path.normpath(os.path.abspath(new_wd))
        self._old_wd = None
    def __enter__(self):
        self._old_wd = os.getcwd()
        os.chdir(self._new_wd)
    def __exit__(self, exc_type, exc_val, exc_tb):
        # @contextmanager versions do not
        # run statements after yield on failure
        os.chdir(self._old_wd)
        self._old_wd = None


def path_from_root(*parts):
    # we dont use os.getcwd in case the wd has changed!
    root = os.path.dirname(os.path.dirname(__file__))
    root = os.path.normpath(os.path.join(root, *parts))
    return root


def read_from_root(*parts: str) -> str:
    with open(path_from_root(*parts), 'r') as f:
        return f.read()


def to_import_path(path: str):
    assert path.endswith('.py')
    path = path[:-len('.py')]
    path = os.path.abspath(os.path.normpath(path))
    path = os.path.relpath(path, os.getcwd())
    return '.'.join(path.split('/'))



import keyword as _keyword

# ========================================================================= #
# Keys                                                                      #
# ========================================================================= #


# keys - all

KEY_TYPE = '__type__'
# keys - options
KEY_PKG = '__package__'
KEY_DEFAULTS = '__defaults__'
KEY_DATA = '__data__'
# keys - groups
KEY_CHILDREN = '__children__'

# marker keys - these are not reserved
MARKER_KEY_TARGET = '_target_'
MARKER_KEY_ARGS   = '_args_'
MARKER_KEY_KWARGS = '_kwargs_'
MARKER_KEY_VALUE  = '_value_'   # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED

# special types
TYPE_OPTION = 'option'
TYPE_GROUP  = 'group'
TYPE_NODE   = 'node'            # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED

# special packages
PKG_ROOT = '<root>'
PKG_GROUP = '<group>'
PKG_OPTION = '<option>'

ALL_PACKAGE_ALIASES = {
    PKG_ROOT,
    PKG_GROUP,
    PKG_OPTION,
}

# special options
OPT_SELF = '<self>'
OPT_GLOB = '*'

ALL_OPT_ALIASES = {
    OPT_SELF, OPT_GLOB
}

# default
DEFAULT_PKG = PKG_GROUP

# - - - - - - - - - - - - - #
# all values                #
# - - - - - - - - - - - - - #

# these keys are allowed as values anywhere and are not reserved
_MARKER_KEYS = {
    MARKER_KEY_TARGET,
    MARKER_KEY_ARGS,
    MARKER_KEY_KWARGS,
    MARKER_KEY_VALUE
}

RESERVED_OPTION_KEYS = {
    KEY_TYPE, KEY_PKG, KEY_DEFAULTS, KEY_DATA,
}

RESERVED_GROUP_KEYS = {
    KEY_TYPE, KEY_CHILDREN,
}

# TODO: not yet implemented
_RESERVED_NODE_KEYS = {
    KEY_TYPE, KEY_DATA,
}

# these keys are not allowed in normal data
RESERVED_KEYS = {
    *RESERVED_GROUP_KEYS,
    *RESERVED_OPTION_KEYS,
    *_RESERVED_NODE_KEYS,
}


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from glob import glob
import os
import ruamel.yaml as yaml

from eunomia.backend import Backend
from eunomia.backend._backend_dict import normalise_option_dict, BackendDict
from eunomia.config import Group, Option
from eunomia.config.nodes import IgnoreNode


# ========================================================================= #
# Config Backend                                                            #
# ========================================================================= #


class BackendYaml(Backend):

    GROUP_TYPE = str
    OPTION_TYPE = str

    def _get_sorted_option_paths(self, root_folder: str):
        ext = '.yaml'
        # get all yaml files
        found_paths = glob(os.path.join(root_folder, f'**/*{ext}'), recursive=True)
        # remove root folder from names
        paths = []
        for path in found_paths:
            # get the relative path, strip the extension, and split
            p = os.path.relpath(path, root_folder)
            p = os.path.splitext(p)[0]
            p = _os_split_path(p)
            # append the path and split path used to sort
            paths.append((path, p))
        # sort paths by length, then alphabetically
        return sorted(paths, key=lambda p: (len(p[1]), *p[1]))

    def _load_group(self, value: GROUP_TYPE) -> Group:
        if not os.path.isdir(value):
            raise FileNotFoundError(f'root_folder={repr(value)} is not a valid directory.\n\tAre you sure the path is correct and not relative?\n\tCurrent working directory is: {repr(os.getcwdb().decode())}')
        # load group
        root = Group()
        for path, (*subgroups, option_name) in self._get_sorted_option_paths(value):
            # add subgroups
            group = root.get_group_recursive(subgroups, make_missing=True)
            # add option by loading file
            group.add_option(option_name, self.load_option(path))
        # done!
        return root

    def _load_option(self, value: OPTION_TYPE) -> Option:
        if not os.path.isfile(value):
            raise FileNotFoundError(f'root_folder={repr(value)} is not a file.\n\tAre you sure the path is correct and not relative?\n\tCurrent working directory is: {repr(os.getcwdb().decode())}')
        if os.path.splitext(value)[1] != '.yaml':
            raise ValueError(f'option file has incorrect extension: {repr(os.path.splitext(value)[1])}, should be a .yaml file')
        # load option
        data = yaml_load_file(value)
        # convert data to option
        return BackendDict(allow_compact_load=True).load_option(data)

    def _dump_group(self, group: Group):
        raise RuntimeError('Not implemented!')  # pragma: no cover

    def _dump_option(self, option: Option):
        raise RuntimeError('Not implemented!')  # pragma: no cover


# ========================================================================= #
# Path Util                                                                 #
# ========================================================================= #


def _os_split_path(path, force_relative=True):
    # get the path
    if force_relative:
        assert not os.path.isabs(path)
    # get the normalised path
    path = os.path.normpath(path)
    # split the path into its keys
    parts, head = [], path
    while True:
        head, tail = os.path.split(head)
        if tail:
            parts.append(tail)
            continue
        elif head:
            parts.append(head)
        break
    parts = parts[::-1]
    # check the path is the original
    recon_path = os.path.join(*parts) if parts else ''
    if recon_path != path:
        raise RuntimeError(f'reconstructed path {repr(recon_path)} is not equal to original {repr(path)}')
    return parts


# ========================================================================= #
# Yaml Loader                                                               #
# ========================================================================= #


class EunomiaSafeLoader(yaml.SafeLoader):

    def __init__(self, *args, always_substitute_strings=True, **kwargs):
        super().__init__(*args, **kwargs)
        # custom config values
        self._always_substitute_strings = always_substitute_strings

    @classmethod
    def add_constructors(cls, tags: list, constructor):
        for tag in tags:
            cls.add_constructor(tag, constructor)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Custom Constructors                                                   #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # def construct_custom_tuple(self, node: yaml.Node):
    #     if not isinstance(node, yaml.SequenceNode):
    #         raise yaml.YAMLError(f'tag {node.tag} for tuple(...) is not compatible with node: {node.__class__.__name__}. {repr(node.value)} must be a sequence/list')
    #     return tuple(self.construct_sequence(node))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Node Constructors                                                     #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def construct_node_ignore(self, node: yaml.Node):
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError(f'tag {node.tag} for {IgnoreNode.__name__} is not compatible with node: {node.__class__.__name__} which is not a scalar.')
        return IgnoreNode(self.construct_scalar(node))

    # def construct_node_ref(self, node: yaml.Node):
    #     return RefNode(self.construct_yaml_str(node))

    # def construct_node_eval(self, node: yaml.Node):
    #     return EvalNode(self.construct_yaml_str(node))

    # def construct_node_sub(self, node: yaml.Node):
    #     return SubNode(self.construct_yaml_str(node))


# EunomiaSafeLoader.add_constructors(['!tuple'], EunomiaSafeLoader.construct_custom_tuple)
EunomiaSafeLoader.add_constructors(['!str'], EunomiaSafeLoader.construct_node_ignore)
# EunomiaSafeLoader.add_constructors(['!ref'], EunomiaSafeLoader.construct_node_ref)
# EunomiaSafeLoader.add_constructors(['!expr'], EunomiaSafeLoader.construct_node_eval)
# EunomiaSafeLoader.add_constructors(['!sub'], EunomiaSafeLoader.construct_node_sub)


# NOTE: unknown tags can be parsed
#       - https://github.com/kislyuk/yq/blob/2dd5cf39e18f2caf0c416e3da1c93a62bc801e0b/yq/loader.py#L57
# TODO: hydra adds the following implicit_resolvers & filters them
#       - float: https://github.com/omry/omegaconf/blob/ad593a527556583dfefa6d7dd34beec3f183a438/omegaconf/_utils.py#L106
#       - timestamp: https://github.com/omry/omegaconf/blob/ad593a527556583dfefa6d7dd34beec3f183a438/omegaconf/_utils.py#L120


# ========================================================================= #
# Yaml Util                                                                 #
# ========================================================================= #


def yaml_load(string):
    return yaml.load(string, EunomiaSafeLoader, version='1.2')


def yaml_load_file(path):
    with open(path, 'r') as f:
        return yaml_load(f.read())


# ========================================================================= #
# End                                                                       #
# ========================================================================= #





# ========================================================================= #
# SimpleAttrDict                                                            #
# ========================================================================= #


class AttrDict(dict):
    """
    Nested Attribute Dictionary

    A class to convert a nested Dictionary into an object with key-values
    accessible using attribute notation (AttrDict.attribute) in addition to
    key notation (Dict["key"]). This class recursively sets Dicts to objects,
    allowing you to recurse into nested dicts (like: AttrDict.attr.attr)

    https://stackoverflow.com/a/48806603/9852408
    """

    def __init__(self, mapping=None):
        super(AttrDict, self).__init__()
        # add all values to dictionary
        if mapping is not None:
            for key, value in mapping.items():
                self.__setitem__(key, value)

    def __setitem__(self, key, value):
        # recursively convert dicts to attrdicts TODO: convert sub-lists and dicts too!
        if isinstance(value, dict):
            value = AttrDict(value)
        # set item like usual
        super(AttrDict, self).__setitem__(key, value)
        # for code completion in editors
        self.__dict__[key] = value

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError

    __setattr__ = __setitem__

    # def __setattr__(self, key, value):
    #     # allow double underscores to be assigned to the class itself
    #     if str.startswith(key, f'_{self.__class__.__name__}__'):
    #         super().__setattr__(key, value)
    #     else:
    #         return self.__setitem__(key, value)



# ========================================================================= #
# End                                                                       #
# ========================================================================= #




from typing import Tuple, List, Union, Optional

from eunomia.config._default import Default
from eunomia.util._util_dict import recursive_getitem, dict_recursive_update
from eunomia.config import Option, Group
from eunomia.config.nodes import ConfigNode

from eunomia.config import keys as K
from eunomia.config import validate as V


# ========================================================================= #
# Config Loader                                                             #
# ========================================================================= #


class ConfigLoader(object):

    def __init__(self, root_group: Group, overrides: list = None):
        # check root group
        if not isinstance(root_group, Group):
            raise TypeError(f'root_group must be a {Group.__name__}')
        self._root_group = root_group
        # has_run
        self._has_run = False
        # merged items
        self._merged_options = {}
        self._first_merge_from = {}
        self._merged_config = {}
        # make defaults overrides
        self._overrides = self._resolve_overrides_list(overrides)
        self._overridden = {}
        # debug
        self._debug = False
        self._debug_depth = 0

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Checks                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _pre_merge_checks(self):
        # 2.1 check all overrides exist in config
        # NOT NEEDED BECAUSE OF OVERRIDE RESOLVER
        # if self._overrides:
        #     for group_keys, (o_grp, o_opts) in self._overrides.items():
        #         for opt in o_opts:
        #             if not self._root_group.has_option_recursive(group_keys + (opt.key,)):
        #                 raise KeyError(f'specified override does not exist in the config: {V.keys_as_abs_config_path(group_keys + (opt.key,))}')
        pass

    def _post_merge_checks(self):
        # ===================== #
        # all values have not yet been resolved in the merge config!
        # ===================== #
        # 2.3 make sure all overrides were used!
        if self._overrides:
            unused_overrides = set(self._overrides.keys()) - set(self._overridden.keys())
            if unused_overrides:
                raise RuntimeError(
                    f'the following overrides were not used to override defaults listed in the config: {sorted(map(V.keys_as_abs_config_path, unused_overrides))}')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Core Algorithm                                                        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def load_config(self, config_name, return_merged_options=False):
        """
        flatten and merge the options lists using DFS, while
        simultaneously merging the config

        - When config dictionaries are encountered, those are also
          processed using DFS to obtain substituted values, before
          being merged into the config.
            * Keys are not allowed to be substituted values
        """
        # ===================== #
        # make sure we cant run the loader more than once
        if self._has_run:
            raise RuntimeError('Loader cannot be run more than once!')
        self._has_run = True
        # ===================== #
        # 1. entry point for dfs, get initial option
        entry_option = self._root_group.get_option(config_name)
        if self._debug:
            print(f'\nentrypoint: {repr(entry_option.abs_path)}')
        # ===================== #
        # 2.1. pre checks
        self._pre_merge_checks()
        # 2.2. perform dfs & merging and finally resolve values
        self._visit_option(entry_option, parent_option=None)
        # 2.3 post checks
        self._post_merge_checks()
        # ===================== #
        # 3. finally resolve all the values in the config
        self._resolve_all_values()
        # ===================== #

        # done, return the result
        if return_merged_options:
            return self._merged_config, self._merged_options

        return self._merged_config

    def _visit_option(self, option: Option, parent_option: Optional[Option]):
        if self._debug:
            print(f'{" "*self._debug_depth*4}> visiting {repr(option.abs_path)} from {repr(parent_option.abs_path) if parent_option else "<user>"}')
            self._debug_depth += 1
        # ===================== #
        # 1. check where to process self, and make sure self is in the options list
        # by default we want to merge this before children so we can reference its values.
        handled_defaults = set()
        defaults = option.get_unresolved_defaults()
        if not any(default.is_self_pre_resolve for default in defaults):
            # allow referencing parent values in children
            defaults = [Default.make_self()] + defaults
            # allow parent overwriting children values
            # group_paths = group_paths + [s.OPT_SELF]
        # ===================== #
        # 2. process options in order
        for default in defaults:
            # normalise the default
            # -- various different objects as well as substitution need to be handled
            d_obj, d_options, d_pkg_keys, d_is_self = default.to_resolved_components(option, resolver=self._resolve_value)
            if self._debug:
                print(f'{" " * self._debug_depth * 4}? resolved default { {d_obj.abs_path: [o.key for o in d_options]} } from {default._default}')
            # ===================== #
            # make sure that self is only ever handled once
            key = K.OPT_SELF if d_is_self else V.keys_as_abs_config_path(d_obj.keys)
            if key in handled_defaults:
                raise RuntimeError(f'resolved default {key} was encountered more than once in option: {option.abs_path}')
            handled_defaults.add(key)
            # ===================== #
            # handle different cases
            if d_is_self:
                # ===================== #
                # 2.a if self is encountered, merge into config. We skip the
                #     value of the option_name here as it is not needed.
                self._merge_option(*d_options, pkg_keys=d_pkg_keys, parent_option=parent_option)
                # ===================== #
            else:
                # ===================== #
                # allow groups/options to be overridden
                o_obj, o_options = self._resolve_override(option, d_obj, d_options)
                if o_obj is not d_obj:
                    raise AssertionError('this should never happen. overridden default options come from different group.')
                # ===================== #
                # 2.b dfs through options
                # visit the next option
                for o_option in o_options:
                    self._visit_option(o_option, parent_option=option)
                # ===================== #
        if self._debug:
            self._debug_depth -= 1

    def _merge_option(self, option: Option, pkg_keys: Tuple[str], parent_option: Optional[Option]):
        # TODO: this should have different modes
        #       - do not allow from the same group to be merged
        #       - allow from the same group to be merged
        #       - only replace existing values on lhs, don't merge
        #       - only add missing values on lhs, don't merge or replace
        # ===================== #
        # 1. check that the group that the option belongs to has not already been merged.
        #    Then mark the group the option belongs to as visited
        #    - if the same source is still merging an option, then we allow
        #      it even if there are multiple
        # ===================== #
        # the path to the group containing the option being merged
        group_keys = option.group_keys
        # the parent option can merge one or more options from the above group
        # once a parent option has been set, being the first to encounter a default, a
        # different parent option cannot merge options from the same group.
        first_merge_from = self._first_merge_from.setdefault(group_keys, parent_option if parent_option else None)  # none means the user!
        visitor = "<user>" if first_merge_from is None else repr(first_merge_from.abs_group_path)
        if first_merge_from is not parent_option:
            raise RuntimeError(f'The first merge of options from the group {repr(option.abs_group_path)} occurred '
                               f'from {visitor} a different parent option cannot request a merge: {repr(option.abs_group_path)}')
        # check that this exact option has not been visited before
        if group_keys in self._merged_options:
            if option.key in self._merged_options[group_keys]:
                raise KeyError(f'option {repr(option.abs_path)} was previously visited by {visitor}')
        # mark group as visited
        self._merged_options.setdefault(group_keys, []).append(option.key)
        # ===================== #
        # 2. merged the data
        # ===================== #
        # 2.1. get the root config object according to the package
        root = recursive_getitem(self._merged_config, pkg_keys, make_missing=True)
        # 2.2. merge the option into the config
        data = option.get_unresolved_data()
        dict_recursive_update(left=root, right=data, allow_overwrite=True)
        # ===================== #
        if self._debug:
            print(f'{" "*self._debug_depth*4}* merged data from {repr(option.abs_path)} into output config at {pkg_keys}')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Resolving Values                                                      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _resolve_override(self, option: Option, d_obj: Union[Group, Option], d_options: List[Option]) -> Tuple[Union[Group, Option], List[Option]]:
        # if there is no listed override, return the usual options to merge
        d_group_keys = d_obj.group_keys
        if d_group_keys not in self._overrides:
            return d_obj, d_options
        # make sure it has not been overridden before
        if d_group_keys in self._overridden:
            raise RuntimeError(f'default lists entry for group: {option.abs_group_path} has already been overridden: "{d_obj.abs_group_path}: {tuple(o.key for o in d_options)}"')
        self._overridden[d_group_keys] = (d_obj, d_options)
        # if there is an override, return the override
        return self._overrides[d_group_keys]

    def _resolve_overrides_list(self, overrides: list) -> dict:
        if overrides is None:
            return {}
        # resolve overrides
        overrides_dct = {}
        for override in overrides:
            o = Default(override)
            if o.is_self_pre_resolve:
                raise KeyError(f'override cannot be self: {repr(override)}')
            try:
                o_obj, o_options, _, o_is_self = o.to_resolved_components(self._root_group, resolver=None, force_substitute=False)
            except KeyError as e:
                raise KeyError(f'could not resolve the override {repr(override)}, are you sure it exists in the config tree?').with_traceback(e.__traceback__)
            if o_is_self:
                raise KeyError(f'override cannot resolve to self: {repr(override)}')
            overrides_dct[o_obj.group_keys] = (o_obj, o_options)
        # return!
        return overrides_dct

    def _resolve_value(self, value):
        # 1. allow interpolation of config objects
        # 2. TODO: process dictionary syntax with _node_ keys
        if isinstance(value, ConfigNode):
            value = value.get_config_value(self._merged_config, self._merged_options, {})
        return value

    def _resolve_all_values(self):
        # TODO: this is wrong!
        # TODO: config is not mutated on the fly, cannot substitute chains of values.
        self._merged_config = ConfigNode.recursive_get_config_value(
            self._merged_config,
            self._merged_options,
            {},
            self._merged_config
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # END Loader                                                            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import pytest
from eunomia.config import keys as K, Option, Group
from eunomia.config import validate as V

# ========================================================================= #
# Test YAML & Custom Tags                                                   #
# ========================================================================= #


def _do_identifier_tests(validator, prefix=''):
    # check simple valid
    validator(prefix + 'valid_id')
    with pytest.raises(ValueError):  # TODO: fix , match="is_not_special_key"):
        validator(prefix + '__type__')
    with pytest.raises(ValueError):  # TODO: fix , match="is_not_reserved"):
        validator(prefix + '__invalid_id__')
    validator(prefix + '_okay_id_')
    with pytest.raises(ValueError):  # TODO: fix , match="is_identifier"):
        validator(prefix + '1234')
    with pytest.raises(ValueError):  # TODO: fix , match="is_not_keyword"):
        validator(prefix + 'with')
    with pytest.raises(TypeError):  # TODO: fix , match="should be instance of"):
        validator(1234)


def test_identifier():
    _do_identifier_tests(V.validate_config_identifier)


def test_pkg_path():
    # check - absolute
    assert V.validate_package_path('') == ''
    assert V.validate_package_path('valid1') == 'valid1'
    assert V.validate_package_path('valid1.valid2') == 'valid1.valid2'
    with pytest.raises(ValueError): V.validate_package_path('valid1.')
    # check - relative
    assert V.validate_package_path('.') == '.'
    assert V.validate_package_path('.valid1') == '.valid1'
    assert V.validate_package_path('.valid1.valid2') == '.valid1.valid2'
    with pytest.raises(ValueError): V.validate_package_path('.valid1.')

    # check splits - absolute
    assert V.split_package_path('') == ([], False)
    assert V.split_package_path('valid1') == (['valid1'], False)
    assert V.split_package_path('valid1.valid2') == (['valid1', 'valid2'], False)
    with pytest.raises(ValueError): V.split_package_path('valid1.')
    # check splits - relative
    assert V.split_package_path('.') == ([], True)
    assert V.split_package_path('.valid1') == (['valid1'], True)
    assert V.split_package_path('.valid1.valid2') == (['valid1', 'valid2'], True)
    with pytest.raises(ValueError): V.split_package_path('.valid1.')

    # check identifiers, should be same as above
    _do_identifier_tests(V.validate_package_path)
    _do_identifier_tests(V.validate_package_path, prefix='valid1.')
    _do_identifier_tests(V.validate_package_path, prefix='_marker_.')
    with pytest.raises(ValueError):  # TODO: fix , match='is_not_reserved'):
        _do_identifier_tests(V.validate_package_path, prefix='__invalid1__.')

    # check group...
    with pytest.raises(ValueError):  # TODO: fix , match='is_identifier'):
        V.validate_package_path('valid1/valid2/valid3')

    # check special keys
    V.validate_package_path('<root>')
    V.validate_package_path('<group>')
    # split special keys
    with pytest.raises(RuntimeError): V.split_package_path('<root>')
    with pytest.raises(RuntimeError): V.split_package_path('<group>')


def test_group_path():
    # check - relative
    assert V.validate_config_path('') == ''
    assert V.validate_config_path('valid1') == 'valid1'
    assert V.validate_config_path('valid1/valid2') == 'valid1/valid2'
    with pytest.raises(ValueError): V.validate_config_path('valid1/')
    # check - absolute
    assert V.validate_config_path('/') == '/'
    assert V.validate_config_path('/valid1') == '/valid1'
    assert V.validate_config_path('/valid1/valid2') == '/valid1/valid2'
    with pytest.raises(ValueError): V.validate_config_path('/valid1/')

    # check splits - relative
    assert V.split_config_path('') == ([], True)
    assert V.split_config_path('valid1') == (['valid1'], True)
    assert V.split_config_path('valid1/valid2') == (['valid1', 'valid2'], True)
    with pytest.raises(ValueError): V.split_config_path('valid1/')
    # check splits - absolute
    assert V.split_config_path('/') == ([], False)
    assert V.split_config_path('/valid1') == (['valid1'], False)
    assert V.split_config_path('/valid1/valid2') == (['valid1', 'valid2'], False)
    with pytest.raises(ValueError): V.split_config_path('/valid1/')

    # check identifiers, should be same as above
    _do_identifier_tests(V.validate_config_path)
    _do_identifier_tests(V.validate_config_path, prefix='valid1/')
    _do_identifier_tests(V.validate_config_path, prefix='_marker_/')
    with pytest.raises(ValueError):  # TODO: fix , match='is_not_reserved'):
        _do_identifier_tests(V.validate_config_path, prefix='__invalid1__/')

    # check group...
    with pytest.raises(ValueError):  # TODO: fix , match='is_identifier'):
        V.validate_config_path('valid1.valid2.valid3')

    # check special keys
    with pytest.raises(ValueError): V.validate_config_path('<root>')
    with pytest.raises(ValueError): V.validate_config_path('<group>')
    # split special keys
    with pytest.raises(RuntimeError): V.split_package_path('<root>')
    with pytest.raises(RuntimeError): V.split_package_path('<group>')


def test_defaults():

    # check self
    V.validate_defaults_item('<self>')
    with pytest.raises(ValueError):
        V.validate_defaults_item('<self>s')

    # check strings
    V.validate_defaults_item('option1')
    V.validate_defaults_item('/option1')
    V.validate_defaults_item('group/option1')
    V.validate_defaults_item('/group/option1')
    with pytest.raises(ValueError):
        V.validate_defaults_item('/1invalidgroup/option1')

    # check dictionaries
    V.validate_defaults_item({'': 'option1'})
    V.validate_defaults_item({'/': 'option1'})
    V.validate_defaults_item({'group': 'option1'})
    V.validate_defaults_item({'/group': 'option1'})
    with pytest.raises(ValueError):
        V.validate_defaults_item({'/group/subgroup': '1invalidopt'})
    with pytest.raises(ValueError):
        V.validate_defaults_item({'/1invalidgroup/subgroup': 'option1'})

    # check tuples
    V.validate_defaults_item({'': 'option1'})
    V.validate_defaults_item({'/': 'option1'})
    V.validate_defaults_item({'group': 'option1'})
    V.validate_defaults_item({'/group': 'option1'})

    with pytest.raises(TypeError):
        V.validate_defaults_item(['', 'option1'])
    with pytest.raises(TypeError):
        V.validate_defaults_item(('', 'option1'))

    # check options as refs
    V.validate_defaults_item(Group(), allow_config_nodes=True)
    V.validate_defaults_item(Option(), allow_config_nodes=True)
    with pytest.raises(TypeError, match='disabled as the group key for default entry'):
        V.validate_defaults_item(Group())
    with pytest.raises(TypeError, match='disabled as the group key for default entry'):
        V.validate_defaults_item(Option())

    # check lots
    V.validate_option_defaults([
        'group1/option1',
        '/group2/option2',
        {'/group/subgroup': 'option1'},
        {'group/subgroup2': 'option2'},
        '<self>',
        Option(),
    ], allow_config_nodes=True)

